;; Generated package description from gnome-c-style.el
(define-package "gnome-c-style" "0.1" "minor mode for editing GNOME-style C source code" 'nil :url "http://elpa.gnu.org/packages/gnome-c-style.html" :keywords '("gnome" "c" "coding style"))
